from pydap.model import DatasetType


class ResponseSerializer(object):
    """
    A serializer for responses that keeps the dataset
    for modification by WSGI middleware.

    Follows this specification::

      http://wsgi.org/wsgi/Specifications/avoiding_serialization

    """
    def __init__(self, dataset, serializer):
        self.dataset = dataset
        self.serializer = serializer

    def x_wsgiorg_parsed_response(self, type):
        if type is DatasetType:
            return self.dataset

    def __iter__(self):
        return self.serializer(self.dataset)



from pydap.model import *
from pydap.lib import __dap__
from pydap.responses.dds import dispatch as dds_dispatch
from pydap.responses.lib import ResponseSerializer
from pydap.xdr import DapPacker


class DODSResponse(object):
    def __init__(self, dataset):
        self.dataset = dataset

    def __call__(self, environ, start_response):
        headers = [('Content-description', 'dods_data'),
                   ('XDODS-Server', 'dods/%s' % '.'.join(str(d) for d in __dap__)),
                   ('Content-type', 'application/octet-stream')]

        def serialize(dataset):
            # Generate DDS.
            for line in dds_dispatch(dataset):
                yield line
            yield 'Data:\n'
            for line in DapPacker(dataset):
                yield line
            if hasattr(dataset, 'close'): dataset.close()

        start_response('200 OK', headers)

        if environ.get('x-wsgiorg.want_parsed_response'):
            return ResponseSerializer(self.dataset, serialize)
        else:
            return serialize(self.dataset)



from pydap.model import *
from pydap.model import typemap
from pydap.lib import __dap__, INDENT, encode_atom, isiterable
from pydap.responses.lib import ResponseSerializer


class DASResponse(object):
    def __init__(self, dataset):
        self.dataset = dataset

    def __call__(self, environ, start_response):
        headers = [('Content-description', 'dods_das'),
                   ('XDODS-Server', 'dods/%s' % '.'.join(str(d) for d in __dap__)),
                   ('Content-type', 'text/plain')]

        def serialize(dataset):
            output = ''.join(dispatch(dataset))
            if hasattr(dataset, 'close'): dataset.close()
            return [output]

        start_response('200 OK', headers)

        if environ.get('x-wsgiorg.want_parsed_response'):
            return ResponseSerializer(self.dataset, serialize)
        else:
            return serialize(self.dataset)


def dispatch(var, level=0):
    dispatchers = [(DatasetType, _dataset),
                   (GridType, _base),
                   (StructureType, _structure),
                   (BaseType, _base)]

    for klass, func in dispatchers:
        if isinstance(var, klass):
            return func(var, level)


def _dataset(var, level=0):
    yield '%sAttributes {\n' % (level * INDENT)

    for attr, values in var.attributes.items():
        for line in _recursive_build(attr, values, level):
            yield line

    for child in var.walk():
        for line in dispatch(child, level=level+1):
            yield line
    yield '%s}\n' % (level * INDENT)


def _structure(var, level=0):
    yield '%s%s {\n' % (level * INDENT, var.name)

    for attr, values in var.attributes.items():
        for line in _recursive_build(attr, values, level):
            yield line

    for child in var.walk():
        for line in dispatch(child, level=level+1):
            yield line
    yield '%s}\n' % (level * INDENT)


def _base(dapvar, level=0):
    yield '%s%s {\n' % (level * INDENT, dapvar.name)

    for attr, values in dapvar.attributes.items():
        for line in _recursive_build(attr, values, level):
            yield line
    yield '%s}\n' % (level * INDENT)


def _recursive_build(attr, values, level=0):
    """
    Recursive function to build the DAS.
    
    """
    # Check for metadata.
    if isinstance(values, dict):
        yield '%s%s {\n' % ((level+1) * INDENT, attr)
        for k, v in values.items():
            for line in _recursive_build(k, v, level+1):
                yield line
        yield '%s}\n' % ((level+1) * INDENT)
    else:
        atoms = values
        if not isiterable(atoms):
            atoms = [atoms]

        # Get value type and encode properly.
        if hasattr(values, 'dtype'):
            type_ = typemap[values.dtype.char].descriptor  # numpy.array
        elif hasattr(values, 'typecode'):
            type_ = typemap[values.typecode].descriptor    # array.array
        else:
            types = [typeconvert(atom) for atom in atoms]
            precedence = ['String', 'Float64', 'Int32']
            types.sort(key=precedence.index)
            type_ = types[0]

        # Encode values
        atoms = [encode_atom(atom) for atom in atoms]
        yield '%s%s %s %s;\n' % ((level+1) * INDENT, type_,
                attr.replace(' ', '_'), ', '.join(atoms))


def typeconvert(obj):
    """Type conversion between Python and DODS, for the DAS."""
    types = [(basestring, 'String'),
             (float, 'Float64'),
             (long, 'Int32'),
             (int, 'Int32')]
    for klass, type_ in types:
        if isinstance(obj, klass):
            return type_



import itertools

import numpy

from pydap.model import *
from pydap.lib import __dap__, INDENT, encode_atom, n_iterate
from pydap.responses.dds import dispatch as dds_dispatch
from pydap.responses.lib import ResponseSerializer


class ASCIIResponse(object):
    def __init__(self, dataset):
        self.dataset = dataset

    def __call__(self, environ, start_response):
        headers = [('Content-description', 'dods_ascii'),
                   ('XDODS-Server', 'dods/%s' % '.'.join(str(d) for d in __dap__)),
                   ('Content-type', 'text/plain')]

        def serialize(dataset):
            # Generate DDS.
            for line in dds_dispatch(dataset):
                yield line
            yield 45 * '-'
            yield '\n'
            for line in dispatch(dataset):
                yield line
            if hasattr(dataset, 'close'): dataset.close()

        start_response('200 OK', headers)

        if environ.get('x-wsgiorg.want_parsed_response'):
            return ResponseSerializer(self.dataset, serialize)
        else:
            return serialize(self.dataset)


def dispatch(var, printname=True):
    dispatchers = [(SequenceType, _sequence),
                   (StructureType, _structure),
                   (BaseType, _base)]
    
    for klass, func in dispatchers:
        if isinstance(var, klass):
            return func(var, printname)


def _structure(var, printname):
    for child in var.walk():
        for line in dispatch(child, printname):
            yield line
        yield '\n'


def _sequence(var, printname):
    yield ', '.join(child.id for child in var.values())
    yield '\n'
    for struct_ in var:
        out = []
        for child in struct_:
            for line in dispatch(child, printname=False):
                out.append(line)
        out = ', '.join(out).split('\n')
        for line in out:
            if line:
                yield line.strip(', ') + '\n'


def _base(var, printname):
    if printname:
        yield var.id
        yield '\n'

    data = numpy.asarray(var.data).flat
    for indexes, value in itertools.izip(n_iterate(var.shape), data):
        index = '[%s]' % ']['.join(str(idx) for idx in indexes)
        yield '%s %s\n' % (index, encode_atom(value))
    if not var.shape:
        yield encode_atom(var.data)



from pydap.lib import __dap__, __version__


class VersionResponse(object):
    def __init__(self, dataset=None):
        if hasattr(dataset, 'close'): dataset.close()

    def __call__(self, environ, start_response):
        headers = [('XDODS-Server', 'dods/%s' % '.'.join(str(d) for d in __dap__)),
                   ('Content-type', 'text/plain')]

        output = """Core version: dods/%s
Server version: pydap/%s
""" % ('.'.join(str(d) for d in __dap__), '.'.join(str(d) for d in __version__))

        start_response('200 OK', headers)
        return [output]




from paste.request import construct_url

from pydap.lib import __dap__, __version__
from pydap.util.template import GenshiRenderer, StringLoader
from pydap.responses.lib import ResponseSerializer


DEFAULT_TEMPLATE = """<html xmlns="http://www.w3.org/1999/xhtml">
    <head>
        <title>pyDAP help page for file $location</title>

        <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
    </head>

    <body>
        <h1>How to access this file</h1>

        <p>The resource <code>$location</code> is a dataset available through the <a href="http://opendap.org/">OPeNDAP</a> protocol. The best way to access this dataset is using an <a href="http://www.opendap.org/faq/whatClients.html">OPeNDAP client</a> like Ferret, GrADS or Matlab.</p>

        <p>In case you <strong>don't have</strong> a client installed, you can try to download data by accessing the <a href="${location}.html">web interface</a> for this dataset.</p>

        <hr />
        <p><em><a href="http://pydap.org/">pydap/$version</a></em> &copy; Roberto De Almeida</p>
    </body>
</html>"""


class HelpResponse(object):

    renderer = GenshiRenderer(
            options={}, loader=StringLoader( {'help.html': DEFAULT_TEMPLATE} ))

    def __init__(self, dataset=None):
        self.dataset = dataset

    def __call__(self, environ, start_response):
        headers = [('XDODS-Server', 'dods/%s' % '.'.join(str(d) for d in __dap__)),
                   ('Content-type', 'text/html')]

        def serialize(dataset):
            location = construct_url(environ, with_query_string=False)
            dataset, response = location.rsplit('.', 1)
            context = {
                'location': dataset,
                'version' : '.'.join(str(d) for d in __version__),
            }
            renderer = environ.get('pydap.renderer', self.renderer)
            template = renderer.loader('help.html')
            output = renderer.render(template, context)
            if hasattr(dataset, 'close'): dataset.close()
            return [output]

        start_response('200 OK', headers)

        if environ.get('x-wsgiorg.want_parsed_response'):
            return ResponseSerializer(self.dataset, serialize)
        else:
            return serialize(self.dataset)



"""
A simple file-based Opendap server.

Serves files from a root directory, handling those recognized by installed
handlers.

"""

import re
import os
from urllib import unquote

from paste.urlparser import StaticURLParser

from pydap.handlers.lib import get_handler, load_handlers
from pydap.lib import __version__
from pydap.exceptions import ExtensionNotSupportedError
from pydap.util.template import FileLoader, GenshiRenderer


class FileServer(StaticURLParser):
    def __init__(self, root, root_directory=None, cache_max_age=None, templates='templates', **config):
        StaticURLParser.__init__(self, root, root_directory, cache_max_age)
        self.config = config

        loader = FileLoader(templates)
        self.renderer = GenshiRenderer(
                options={}, loader=loader)

        self.handlers = load_handlers()

    def not_found(self, environ, start_response, debug_message=None):
        request = os.path.split(environ['SCRIPT_NAME'])[1]
        request = os.path.normpath(os.path.join(self.directory, request))
        filepath, response = os.path.splitext(request)
        if os.path.sep != '/': filepath = filepath.replace('/', os.path.sep)

        if environ['PATH_INFO'] == '/':
            return self.index(environ, start_response)
        elif os.path.exists(filepath):
            # Update environ with configuration keys (environ wins in case of conflict).
            for k in self.config:
                environ.setdefault(k, self.config[k])
            environ['pydap.renderer'] = self.renderer
            handler = get_handler(filepath, self.handlers)
            return handler(environ, start_response)
        else:
            return StaticURLParser.not_found(
                    self, environ, start_response, debug_message)

    def index(self, environ, start_response):
        # Return directory listing.
        dirs = []
        files = []
        for name in os.listdir(self.directory):
            path = os.path.normpath(os.path.join(self.directory, name))
            if os.path.isdir(path) and not name.startswith('.'):
                dirs.append(name)
            elif os.path.isfile(path) and not name.startswith('.'):
                file = {'name': name,
                        'size': format_size(os.stat(path)[6]),
                        'supported': False}
                try:
                    get_handler(path, self.handlers)
                    file['supported'] = True
                except ExtensionNotSupportedError:
                    pass
                files.append(file)
        # Sort naturally using Ned Batchelder's algorithm.
        dirs.sort(key=alphanum_key)
        files.sort(key=lambda l: alphanum_key(l['name']))

        context = {
                'title': 'Index of %s' % unquote(environ.get('SCRIPT_NAME') or '/'),
                'dirs' : dirs,
                'files': files,
                'version': '.'.join(str(d) for d in __version__)
        }
        template = self.renderer.loader('index.html')
        output = self.renderer.render(template, context)
        headers = [('Content-type', 'text/html; charset=utf-8')]
        start_response("200 OK", headers)
        return [output]


# http://svn.colorstudy.com/home/ianb/ImageIndex/indexer.py
def format_size(size):
    if not size:
        return 'empty'
    if size > 1024:
        size = size / 1024.
        if size > 1024:
            size = size / 1024.
            return '%.1iMb' % size
        return '%.1fKb' % size
    return '%i bytes' % size


def alphanum_key(s):
    """
    Turn a string into a list of string and number chunks.

        >>> alphanum_key("z23a")
        ['z', 23, 'a']

    From http://nedbatchelder.com/blog/200712.html#e20071211T054956

    """
    def tryint(s):
        try:
            return int(s)
        except:
            return s
    return [tryint(c) for c in re.split('([0-9]+)', s)]


def make_app(global_conf, root, templates, **kwargs):
    return FileServer(root, templates=templates, **kwargs)
